// SPRx34 jcv 22-Jul-2010 Enhanced verbosity for leap second filenames handling

#include <sstream>

#include "errstr.h"

#include "DataFiles.h"

DataFiles::DataFiles(string remote, string local, bool s2k) : remoteURL(remote), localDirectory(local),
	isS2k(s2k) {
  
  // All Done.
}

// given a time range, return a time-ordered list of data files to process
std::vector<string> DataFiles::localDataFiles(STime start, STime stop) {

  std::vector<string> theFiles;

  errstr << "local: Data time range is " << start << " to " << stop << Sas::message(Msg::App, Msg::Verbose);

  STime first = whichTime(start);
  STime last = whichTime(stop);

  errstr << "local: File time range is " << first << " to " << last << Sas::message(Msg::App, Msg::Verbose);

  if (first == last) {
    theFiles.push_back(getLocalFilename(first));
  }
  else {
    // FIXME Really ugly, but first <= last is broken
    while ( (first < last) || (first.modifiedJulianDay().full() == last.modifiedJulianDay().full()) ) {
        // SPRx34
        errstr << "Seeking local files for time " << first << Sas::message(Msg::App, Msg::Verbose);
        // END SPRx34
    	theFiles.push_back(getLocalFilename(first));

        // SPRx34 when years with added second, should be 43201 instead of 43200
	//first += 43200;
        first=STime(first.year(),first.month(),first.day(),12+first.hour());

    }
  }

  return theFiles;

}

// Gets the data file for a given time, throwing an exception if there is no file found
string DataFiles::getLocalFilename(STime time){

  string filename;

  std::ostringstream ostr;
  
  ostr << "*" 
       << time.year() << '_'
       << setw(2) << setfill('0') << time.month() << '_' 
       << setw(2) << setfill('0') << time.day() << "[Tt]" 
       << setw(2) << setfill('0') << time.hour() << "*";

  localDirectory.filter(ostr.str());

  int nFiles = 0;

  for(Directory::const_iterator i = localDirectory.begin(); i != localDirectory.end(); i++, nFiles++){
    filename = (*i)->name();
  }
  
  if (nFiles == 0) {
    errstr << "No data file found in local directory for " << time
	   << Sas::fatal("noDataFile");
  }

  // TODO Check for multiple files (non-merged CDs)

  errstr << "Filename in local directory for " << time << " is " << filename << Sas::message(Msg::App, Msg::Verbose);
  return filename;

}

// Given a time, calculate the date of the Data File in which it falls
STime DataFiles::whichTime(STime time) {

  if (time.hour() >= 12) {
    return STime(time.year(),time.month(),time.day(),12);
  }
  else {
    return STime(time.year(),time.month(),time.day());
  }
  
}

string DataFiles::getDirectory() {

  return localDirectory.name();

}

// given a time range, return a time-ordered list of data files to process
std::vector<string> DataFiles::remoteDataFiles(STime start, STime stop) {

  std::vector<string> theFiles;

  errstr << "remote: Data time range is " << start << " to " << stop << Sas::message(Msg::App, Msg::Verbose);

  STime first = whichTime(start);
  STime last = whichTime(stop);

  errstr << "remote: File time range is " << first << " to " << last << Sas::message(Msg::App, Msg::Verbose);

  if (first == last) {
    theFiles.push_back(getRemoteFilename(first));
  }
  else {
    // FIXME Really ugly, but first <= last is broken
    while ( (first < last) || (first.modifiedJulianDay().full() == last.modifiedJulianDay().full()) ) {
        // SPRx34
        errstr << "Seeking remote files for time " << first << Sas::message(Msg::App, Msg::Verbose);
        // END SPRx34
    	theFiles.push_back(getRemoteFilename(first));
        // SPRx34 when years with added second, should be 43201 instead of 43200
	//first += 43200;
        first=STime(first.year(),first.month(),first.day(),12+first.hour());
    }
  }

  return theFiles;

}

// Gets the data file for a given time, throwing an exception if there is no file found
string DataFiles::getRemoteFilename(STime time){

  string filename;

  std::ostringstream ostr;

   if (isS2k) {
	  ostr << remoteURL
    	   << "X2KA_" 
    	   << time.year() << '_'
    	   << setw(2) << setfill('0') << time.month() << '_' 
    	   << setw(2) << setfill('0') << time.day() << "T" 
    	   << setw(2) << setfill('0') << time.hour() << "_00_00Z__.DAT";   
   }
   else {
	  ostr << remoteURL
    	   << "XMCS_"
    	   << time.year() << '_'
    	   << setw(2) << setfill('0') << time.month() << '_'
    	   << setw(2) << setfill('0') << time.day() << "T"
    	   << setw(2) << setfill('0') << time.hour() << "_00_00Z__.DAT";
   }

  errstr << "URL for " << time << " is " << ostr.str() << Sas::message(Msg::App, Msg::Verbose);

  return ostr.str();

}
