// CRx17    jcv 05-May-09 Original 
// CRx18    jcv 07-Jul-09 Incorporated the packet processing 
// SPRx32   jcv 07-May-10 Added idle frames checking
// SPRx33   jcv 24-May-10 Frame Time dump in readable format 
// CRx19   jcv 31-Aug-10 New ground station Kourou-Galliot 
// CRx21   jcv 27-Sep-10 Added hex dump of transfer frame header

#include "TMFrameStream.h"
#include "TMPacketProcessor.h"
#include "DataDefinition.h"
#include "Utils.h"

#include "FilterDefinition.h"

#include "errstr.h"

TMFrameStream::TMFrameStream() {
  
  isFirstFrame = true;
  isOverspill = false;
  iCopied = 0;

  refXMCSCorrection = STime("2002-01-22T12:00:00.0");
  
}


void TMFrameStream::procData(RawTM::XMCS::DataUnit *DU,
                           FilterInfo *Filters,
                           int iNumGeneric, int iNumSpecific,
                           VTMZZ_TNODFHTIME *MyarNoDFHTime )
{
   // This function should be deleted sometime in future
   // it only exist for reverse compatibility with abstract DUProcessor

   
   errstr << "TMFrameStream: Called obsolete function."
         << iNumGeneric << iNumSpecific << Filters->bFilterRequired
         << MyarNoDFHTime[0].uMap.iPktID 
	 << Sas::message(Msg::App, Msg::Verbose);

   // Indeed, it is a synonym
   this->extractData(DU);

}

void TMFrameStream::extractData(RawTM::XMCS::DataUnit *DU)
 {
  
  // Make a copy of the Raw DU data
  if ( DU->Size - RawTM::XMCS::header_size + sizeof(STime) != sizeof(RawTM::NCTRS::TMDU)) {
    errstr << "TMFrameStream: XMM DU with NCTRS DU of wrong size!"
	   << Sas::warning("WrongSize");
    return;
  }    

  ////////////////////////////////////////
  // Performs a memcpy i.e. Original data will NOT be modified
  memcpy(&nctrsDU, DU->Data, DU->Size - RawTM::XMCS::header_size);
  
  // Byte swap NCTRS DU 
  Utils::sunByteSwap(nctrsDU.Header.rawPacketSize, sizeof(nctrsDU.Header.rawPacketSize));
  Utils::sunByteSwap(nctrsDU.Header.rawSCID, sizeof(nctrsDU.Header.rawSCID));
  Utils::sunByteSwap(nctrsDU.Header.rawGSID, sizeof(nctrsDU.Header.rawGSID));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Days, sizeof(nctrsDU.Header.rawTime.Days));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Millis, sizeof(nctrsDU.Header.rawTime.Millis));
  Utils::sunByteSwap(nctrsDU.Header.rawTime.Micros, sizeof(nctrsDU.Header.rawTime.Micros));
  Utils::sunByteSwap(nctrsDU.Header.rawQuality, sizeof(nctrsDU.Header.rawQuality));

  // SPRx33
  STime nctrsDUTime = Utils::cds2STime(nctrsDU.Header.Time);
  // END SPRx33

  errstr << "TMFrameStream: NCTRS Header: " 
	 << " FrameSize " << nctrsDU.Header.PacketSize 
	 << " SCID " << nctrsDU.Header.SCID 
	 << " DSType " << (int)nctrsDU.Header.DSType 
	 << " VCID " << (int)nctrsDU.Header.VCID 
	 << " GSID " << nctrsDU.Header.GSID 
	 << " Quality " << nctrsDU.Header.Quality 
	 << " Time " 
	 << nctrsDU.Header.Time.Days << ":"
	 << nctrsDU.Header.Time.Milliseconds << ":"
	 << nctrsDU.Header.Time.Microseconds             
         // SPRx33
         << " ( " << nctrsDUTime << " ) "
         // SPRx33
	 << Sas::message(Msg::App, Msg::Verbose);

  if (nctrsDU.Header.Quality == RawTM::NCTRS::bad) {

    errstr << "TMFrameStream: Bad quality NCTRS frame"
	   << Sas::warning("BadFrame");
    
    return;
    
  }

  if ( nctrsDU.Header.SCID != 193 ) {
    
    errstr << "TMFrameStream: Data not from S/C: S/C ID = " << nctrsDU.Header.SCID
	   << Sas::warning("BadSC");
    
    return;

  }

  if ( (nctrsDU.Header.GSID != RawTM::NCTRS::VIL1 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::REDU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::VIL2 ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::KRU ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::PER ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::NNO ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::AGO ) &&
       // NEW CRx19
       (nctrsDU.Header.GSID != RawTM::NCTRS::KGAL ) &&
       (nctrsDU.Header.GSID != RawTM::NCTRS::DON ) ) {
    errstr << "TMFrameStream: WARNING Bad GS ID" << nctrsDU.Header.GSID
	   << Sas::warning("BadGS");
    
    return;

  }
  
  // CRx21
  // Print 6 bytes, in hexadecimal
  errstr << "TMFrameStream: NCTRS Transfer Frame (Prim. Header): " 
    << hex
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameID[0]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameID[1]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.MFC
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.VFC
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameStatus[0]
    << setw(2) << setfill ( '0' ) << (int)nctrsDU.Frame.rawFrameStatus[1]
    << dec
    << Sas::message(Msg::App, Msg::Verbose);
  // END CRx21

  // Byte swap Transfer Frame
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameID, sizeof(nctrsDU.Frame.rawFrameID));
  Utils::sunByteSwap(nctrsDU.Frame.rawFrameStatus, sizeof(nctrsDU.Frame.rawFrameStatus));
  Utils::sunByteSwap(nctrsDU.Frame.rawCLCW, sizeof(nctrsDU.Frame.rawCLCW));

  errstr << "TMFrameStream: NCTRS Transfer Frame: " 
	 << " OpControl " << nctrsDU.Frame.FrameID.OpControl     
	 << " VCID " << nctrsDU.Frame.FrameID.VCID	  
	 << " SCID " << nctrsDU.Frame.FrameID.SCID          
	 << " Version " << nctrsDU.Frame.FrameID.Version	  
	 << " MFC " << (int)nctrsDU.Frame.MFC           
	 << " VFC " << (int)nctrsDU.Frame.VFC           
	 << " FirstHeader " << nctrsDU.Frame.FrameStatus.FirstHeader   
	 << " SegID " << nctrsDU.Frame.FrameStatus.SegID	 
	 << " PacketOrder " << nctrsDU.Frame.FrameStatus.PacketOrder   
  	 << " SyncFlag " << nctrsDU.Frame.FrameStatus.SyncFlag      
	 << " SecHeaderFlag " << nctrsDU.Frame.FrameStatus.SecHeaderFlag 
	 << " CLCW " << hex << nctrsDU.Frame.CLCW << dec 
   	 << Sas::message(Msg::App, Msg::Verbose);

  // SPRx32 Check for idle frames (none should ever get here...)
  if (nctrsDU.Frame.FrameStatus.FirstHeader == 0x07FE) {
    errstr << "NCTRS Processor: Idle TM Trame found on VC"
           << (int)nctrsDU.Header.VCID
           << Sas::warning("duIdleFrame");
  }

  ////////////////////////////////////////////////////////
  // Once the frame is identified, the packets are extracted

  // We will need correct the Frame Time for having the packet time
  nctrsDU.CorrectedFrameTime = correctFrameTime(nctrsDU.Header.Time, nctrsDU.Header.GSID);


    RawTM::XMM::PacketHeader pHeader;
    unsigned int iLeft;
    
    unsigned int iOffset = nctrsDU.Frame.FrameStatus.FirstHeader;
    nPackets = 0;
    
    // BEGIN Packet Extraction goes here
    if (isFirstFrame) {
      errstr << "TMFrameStream: Processing first frame" 
             << Sas::message(Msg::App, Msg::Noisy);
      isFirstFrame = false;
    }
    else if (isOverspill) {
      errstr << "TMFrameStream: Overspill to be processed" 
             << Sas::message(Msg::App, Msg::Noisy);
      
      if (nctrsDU.Frame.FrameID.VCID != lastVC) {
        errstr << "TMFrameStream: VC mutated from last frame VC!"
               << Sas::warning("WrongVC");
      }
      else if ( (iOffset + iCopied) <= sizeof( RawTM::XMM::Packet ) ) {
        // Copy the overspill into the new packet
        extractedPackets[nPackets] = overspill;
        // iOffset gives the number of bytes not transmitted in the packet from the last frame

        // THE COPY PROCESS
        memcpy((void *)&extractedPackets[nPackets].Data[iCopied], 
               (void *)&nctrsDU.Frame.Data[0], 
               iOffset);

        memcpy ((void *)&pHeader, 
                (void *)&extractedPackets[nPackets].Data[0], 
                sizeof ( RawTM::XMM::PacketHeader ));

        Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    

        if ( (iCopied + iOffset) == (unsigned int)(pHeader.PacketLength + 7) ) {
          errstr << "TMFrameStream: Overspill processed" 
                 << Sas::message(Msg::App, Msg::Noisy);
          extractedPacketTimes[nPackets] = overspillTime;
          nPackets++;
        }
        else {
          errstr << "TMFrameStream: Overspill mismatch"
                 << Sas::warning("OverspillMismatch");
        }

      }
      else {
        errstr << "TMFrameStream: Discarding Overspill"
               << Sas::warning("OverspillDiscard");
      }
      
    }
    // END Packet Extraction 
    // BEGIN NORMAL PROCESSING
   lastVC = nctrsDU.Frame.FrameID.VCID;
    isOverspill = false;
    iCopied = 0;
    
    bool loop = true;
    bool cont = true;

    while ( loop && cont ) {
      
      iLeft = RawTM::NCTRS::data_size - iOffset;
      
      // Packet header split over frame boundary
      if ( sizeof(RawTM::XMM::PacketHeader) > iLeft ) {

        memcpy((void *)&overspill, 
               (void *)&nctrsDU.Frame.Data[ iOffset ],
               iLeft );

        overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

        loop = false;
        isOverspill = true;
        iCopied = iLeft;

        errstr << "TMFrameStream: Packet Header overspills" 
               << Sas::message(Msg::App, Msg::Noisy);

      }
      else {

        memcpy ((void *)&pHeader, 
                (void *)&nctrsDU.Frame.Data[ iOffset ],
                sizeof ( RawTM::XMM::PacketHeader ));

        Utils::sunByteSwap(pHeader.rawPacketLength, sizeof(pHeader.rawPacketLength));    

        if ( (unsigned int)(pHeader.PacketLength + 7) > iLeft ) {
          
          memcpy((void *)&overspill, 
                 (void *)&nctrsDU.Frame.Data[ iOffset ],
                 iLeft );
          overspillTime = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);

          loop = false;
          isOverspill = true;
          iCopied = iLeft;

          errstr << "TMFrameStream: Packet overspills" 
                 << Sas::message(Msg::App, Msg::Noisy);
          
        }
        else {
          
          memcpy((void *)&extractedPackets[nPackets], 
                 (void *)&nctrsDU.Frame.Data[ iOffset], 
                 pHeader.PacketLength + 7);

          extractedPacketTimes[nPackets] = calculatePacketTime(iOffset, nctrsDU.CorrectedFrameTime);
          
          nPackets++;

          if (nPackets >= RawTM::NCTRS::max_packets) {
            errstr << "TMFrameStream: Too many packets"
                   << Sas::warning("maxPackets");
            cont = false;
          }
          iOffset += pHeader.PacketLength + 7;
          if (iOffset == RawTM::NCTRS::data_size) {
            loop = false;
          }
          errstr << "TMFrameStream: Packet processed" 
                 << Sas::message(Msg::App, Msg::Noisy);
        }

      } // END if packet lenght + 7 > iLeft
      
    } // END while loop and cont 

    errstr << nPackets << " packets extracted from Frame" 
           << Sas::message(Msg::App, Msg::Verbose);

   //////////////// PACKET DATA PROCESSING
   // If there is no registered processor, nothing happens, ok

   for (unsigned int c = 0; c < nPackets; c++) {
    for (unsigned int i = 0 ; i < getMap().size(); i++) {
     TMPacketProcessor *p = (TMPacketProcessor *)getMap()[i];
     p->processPacket( extractedPackets[c], extractedPacketTimes[c]);
    }
   }

   //////////////// END PACKET DATA PROCESSING

 return;

}

STime TMFrameStream::correctFrameTime(Utils::CDSTime Time, unsigned short GSID) {

  STime uncorrectedXMCSFrameTime = Utils::cds2STime(Time);

  double xmcsOBDHDelay = delayHandler.getXMCSOBDHdelay(uncorrectedXMCSFrameTime);
  double xmcsGSDelay   = delayHandler.getXMCSGSdelay(uncorrectedXMCSFrameTime, GSID);
  
  double actualOBDHDelay = delayHandler.getActualOBDHdelay(uncorrectedXMCSFrameTime);
  double actualGSDelay   = delayHandler.getActualGSdelay(uncorrectedXMCSFrameTime, GSID);
  
  // Before 2002-01-22T12:00:00Z, the XMCS overwrote the ERT. Correct for this
  // by reversing the changes
  
  // Frame time, corrected for XMCS overwrite
  STime correctedXMCSFrameTime;
  double signalDelay;
  
  // Calculate signal delay from ERT + XMCS correction (since the XMCS would have 
  // subtracted the GS Delay from the ERT
  if (uncorrectedXMCSFrameTime < refXMCSCorrection) {
    
    STime PropDelayCorrected = uncorrectedXMCSFrameTime + xmcsGSDelay;
    signalDelay = delayHandler.getSignalDelay(PropDelayCorrected, GSID);
    
    correctedXMCSFrameTime = uncorrectedXMCSFrameTime + signalDelay;

    correctedXMCSFrameTime = correctedXMCSFrameTime + xmcsGSDelay;

    correctedXMCSFrameTime = correctedXMCSFrameTime + xmcsOBDHDelay;

  }
  else {
    correctedXMCSFrameTime = uncorrectedXMCSFrameTime;
    signalDelay = delayHandler.getSignalDelay(correctedXMCSFrameTime, GSID);	
  }

  double signalDelay2 = delayHandler.getSignalDelay(correctedXMCSFrameTime - signalDelay, GSID);
    
  // Now add the various delays
  STime correctedFrameTime = correctedXMCSFrameTime;
  
  correctedFrameTime = correctedFrameTime - signalDelay2;
  
  correctedFrameTime = correctedFrameTime - actualGSDelay;

  correctedFrameTime = correctedFrameTime - actualOBDHDelay;

  return correctedFrameTime;

}


STime TMFrameStream::calculatePacketTime(unsigned int iOffset, STime frameTime) {

  STime pktTime = frameTime + ((iOffset * 8) / 70000.0);

  return pktTime;

}

